@@ -27,7 +27,9 @@
 #include <unistd.h>
 #include <string.h>
 #include <signal.h>
+#if HAVE_EXECINFO_H
 #include <execinfo.h>
+#endif
 #include <syslog.h>
 #include <sys/wait.h>
 #include <sys/stat.h>
@@ -50,11 +52,12 @@ typedef struct {
 
 struct GdmSignalHandlerPrivate
 {
-        GMainLoop  *main_loop;
-        GHashTable *lookup;
-        GHashTable *id_lookup;
-        GHashTable *action_lookup;
-        guint       next_id;
+        GHashTable    *lookup;
+        GHashTable    *id_lookup;
+        GHashTable    *action_lookup;
+        guint          next_id;
+        GDestroyNotify fatal_func;
+        gpointer       fatal_data;
 };
 
 static void     gdm_signal_handler_class_init   (GdmSignalHandlerClass *klass);
@@ -140,10 +143,11 @@ signal_io_watch (GIOChannel       *ioc,
         block_signals_pop ();
 
         if (is_fatal) {
-                g_debug ("GdmSignalHandler: Caught termination signal - exiting main loop");
-                if (handler->priv->main_loop != NULL) {
-                        g_main_loop_quit (handler->priv->main_loop);
+                if (handler->priv->fatal_func != NULL) {
+                        g_debug ("GdmSignalHandler: Caught termination signal - calling fatal func");
+                        handler->priv->fatal_func (handler->priv->fatal_data);
                 } else {
+                        g_debug ("GdmSignalHandler: Caught termination signal - exiting");
                         exit (1);
                 }
 
@@ -158,6 +162,7 @@ signal_io_watch (GIOChannel       *ioc,
 static void
 fallback_get_backtrace (void)
 {
+#ifdef HAVE_EXECINFO_H
         void *  frames[64];
         size_t  size;
         char ** strings;
@@ -171,9 +176,10 @@ fallback_get_backtrace (void)
                 }
                 free (strings);
                 syslog (LOG_CRIT, "******************* END **********************************");
-        } else {
-                g_warning ("GDM crashed, but symbols couldn't be retrieved.");
+                return;
         }
+#endif
+        g_warning ("GDM crashed, but symbols couldn't be retrieved.");
 }
 
 
@@ -456,13 +462,14 @@ signal_list_free (GSList *list)
 }
 
 void
-gdm_signal_handler_set_main_loop (GdmSignalHandler *handler,
-                                  GMainLoop        *main_loop)
+gdm_signal_handler_set_fatal_func (GdmSignalHandler *handler,
+                                   GDestroyNotify    func,
+                                   gpointer          user_data)
 {
         g_return_if_fail (GDM_IS_SIGNAL_HANDLER (handler));
 
-        /* FIXME: take a ref */
-        handler->priv->main_loop = main_loop;
+        handler->priv->fatal_func = func;
+        handler->priv->fatal_data = user_data;
 }
 
 static void
@@ -481,6 +488,8 @@ gdm_signal_handler_init (GdmSignalHandler *handler)
         if (pipe (signal_pipes) == -1) {
                 g_error ("Could not create pipe() for signal handling");
         }
+        fcntl(signal_pipes[0], F_SETFD, FD_CLOEXEC);
+        fcntl(signal_pipes[1], F_SETFD, FD_CLOEXEC);
 
         ioc = g_io_channel_unix_new (signal_pipes[0]);
         g_io_channel_set_flags (ioc, G_IO_FLAG_NONBLOCK, NULL);
@@ -519,6 +528,9 @@ gdm_signal_handler_finalize (GObject *object)
         }
         g_hash_table_destroy (handler->priv->action_lookup);
 
+        close (signal_pipes [0]);
+        close (signal_pipes [1]);
+
         G_OBJECT_CLASS (gdm_signal_handler_parent_class)->finalize (object);
 }
 
