@@ -147,7 +147,7 @@ gdm_greeter_session_new (const char *display_name,
         GObject *object;
 
         object = g_object_new (GDM_TYPE_GREETER_SESSION,
-                               "command", LIBEXECDIR "/gdm-simple-greeter",
+                               "command", BINDIR "/gnome-session --autostart="DATADIR"/gdm/autostart/LoginWindow/",
                                "server-dbus-path", GDM_GREETER_SERVER_DBUS_PATH,
                                "server-dbus-interface", GDM_GREETER_SERVER_DBUS_INTERFACE,
                                "server-env-var-name", "GDM_GREETER_DBUS_ADDRESS",
@@ -156,6 +156,7 @@ gdm_greeter_session_new (const char *display_name,
                                "x11-display-device", display_device,
                                "x11-display-hostname", display_hostname,
                                "x11-display-is-local", display_is_local,
+                               "runtime-dir", GDM_SCREENSHOT_DIR,
                                NULL);
 
         return GDM_GREETER_SESSION (object);
